@@ -25,32 +25,29 @@
 # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 # THE SOFTWARE.
 import logging
+import warnings
 from typing import Dict, Iterable, List, Optional, Tuple, Union
 
 import numpy as np
 import pandas as pd
-from component_contribution import DEFAULT_QUILT_PKG as DEFAULT_QUILT_PKG_CC
 from component_contribution import (
-    DEFAULT_QUILT_VERSION as DEFAULT_QUILT_VERSION_CC,
-)
-from component_contribution import (
-    LEGACY_QUILT_VERSION as LEGACY_QUILT_VERSION_CC,
+    ZENODO_DOI_PARAMETERS,
+    ZENODO_DOI_PARAMETERS_LEGACY,
 )
+from component_contribution.linalg import LINALG
 from component_contribution.predict import (
     CCModelParameters,
     GibbsEnergyPredictor,
 )
 from equilibrator_cache import (
+    DEFAULT_ZENODO_DOI,
     PROTON_INCHI_KEY,
     Compound,
     CompoundCache,
     CompoundMicrospecies,
-    create_compound_cache_from_quilt,
-)
-from equilibrator_cache.api import DEFAULT_QUILT_PKG as DEFAULT_QUILT_PKG_CACHE
-from equilibrator_cache.api import (
-    DEFAULT_QUILT_VERSION as DEFAULT_QUILT_VERSION_CACHE,
+    create_compound_cache_from_zenodo,
 )
+from equilibrator_cache.exceptions import MissingDissociationConstantsException
 from equilibrator_cache.reaction import (
     create_stoichiometric_matrix_from_reactions,
 )
@@ -147,11 +144,11 @@ class ComponentContribution(object):
                 "[energy]/[substance]"
             ), "rmse_inf must be in units of kJ/mol (or equivalent)"
 
-        self.ccache = ccache or create_compound_cache_from_quilt()
+        self.ccache = ccache or create_compound_cache_from_zenodo()
 
         if predictor is None:
-            parameters = CCModelParameters.from_quilt(
-                version=DEFAULT_QUILT_VERSION_CC, force=True
+            parameters = CCModelParameters.from_zenodo(
+                zenodo_doi=ZENODO_DOI_PARAMETERS
             )
             self.predictor = GibbsEnergyPredictor(
                 parameters=parameters, rmse_inf=rmse_inf
@@ -222,7 +219,7 @@ class ComponentContribution(object):
         A ComponentContribution object
         """
         cc = ComponentContribution.initialize_custom_version(
-            quilt_version_cc=LEGACY_QUILT_VERSION_CC
+            zenodo_doi_params=ZENODO_DOI_PARAMETERS_LEGACY
         )
         cc.p_mg = Q_(14)  # i.e. set [Mg2+] to (essentially) zero
         return cc
@@ -230,14 +227,8 @@ class ComponentContribution(object):
     @staticmethod
     def initialize_custom_version(
         rmse_inf: Q_ = default_rmse_inf,
-        quilt_package_cache: str = DEFAULT_QUILT_PKG_CACHE,
-        quilt_version_cache: Optional[str] = DEFAULT_QUILT_VERSION_CACHE,
-        quilt_tag_cache: Optional[str] = None,
-        quilt_hash_cache: Optional[str] = None,
-        quilt_package_cc: str = DEFAULT_QUILT_PKG_CC,
-        quilt_version_cc: Optional[str] = DEFAULT_QUILT_VERSION_CC,
-        quilt_tag_cc: Optional[str] = None,
-        quilt_hash_cc: Optional[str] = None,
+        zenodo_doi_cache: str = DEFAULT_ZENODO_DOI,
+        zenodo_doi_params: str = ZENODO_DOI_PARAMETERS,
     ) -> "ComponentContribution":
         """Initialize a ComponentContribution object with custom quilt versions.
 
@@ -248,38 +239,18 @@ class ComponentContribution(object):
             covariance matrix for reactions outside the span of
             Component Contribution.
             (Default value: 1e-5 kJ/mol)
-        quilt_package_cache : str, optional
-            (Default value: "equilibrator/cache")
-        quilt_version_cache : str, optional
-            (Default value: "0.2.10")
-        quilt_tag_cache : str, optional
-        quilt_hash_cache: str, optional
-        quilt_package_cc : str, optional
-            (Default value: "equilibrator/component_contribution")
-        quilt_version_cc : str, optional
-            (Default value: "0.3.0")
-        quilt_tag_cc : str, optional
-        quilt_hash_cc: str, optional
+        zenodo_doi_cache : str, optional
+            (Default value: "10.5281/zenodo.4128543")
+        zenodo_doi_params : str, optional
+            (Default value: "10.5281/zenodo.4013789")
 
         Returns
         -------
         A ComponentContribution object
 
         """
-        ccache = create_compound_cache_from_quilt(
-            package=quilt_package_cache,
-            hash=quilt_hash_cache,
-            tag=quilt_tag_cache,
-            version=quilt_version_cache,
-            force=True,
-        )
-        parameters = CCModelParameters.from_quilt(
-            package=quilt_package_cc,
-            hash=quilt_hash_cc,
-            tag=quilt_tag_cc,
-            version=quilt_version_cc,
-            force=True,
-        )
+        ccache = create_compound_cache_from_zenodo(zenodo_doi_cache)
+        parameters = CCModelParameters.from_zenodo(zenodo_doi_params)
         predictor = GibbsEnergyPredictor(
             parameters=parameters, rmse_inf=rmse_inf
         )
@@ -479,28 +450,46 @@ class ComponentContribution(object):
         return standard_dg + stored_dg
 
     def standard_dg_multi(
-        self, reactions: List[PhasedReaction]
+        self,
+        reactions: List[PhasedReaction],
+        uncertainty_representation: str = "cov",
     ) -> Tuple[np.ndarray, np.ndarray]:
         """Calculate the chemical reaction energies of a list of reactions.
 
         Using the major microspecies of each of the reactants.
 
+        Parameters
+        ----------
+        reactions : List[PhasedReaction]
+            a list of PhasedReaction objects to estimate
+        uncertainty_representation : str
+            which representation to use for the uncertainties. 'cov' would
+            return a full covariance matrix. 'sqrt' would return a sqaure root
+            of the covariance, based on the uncertainty vectors.
+            'fullrank' would return a full-rank square root of the covariance
+            which is a compressed form of the 'sqrt' result.
+            (Default value: 'cov')
+
         Returns
         -------
-        standard_dg : ndarray
-            the CC estimation of the major microspecies' standard formation
-            energy
-        dg_sigma : ndarray
-            the uncertainty matrix.
+        standard_dg : Quantity
+            the estimated standard reaction Gibbs energies based on the the
+            major microspecies
+        dg_uncertainty : Quantity
+            the uncertainty matrix (in either 'cov', 'sqrt' or 'fullrank'
+            format)
         """
         rxn_dg_pairs = map(lambda r: r.separate_stored_dg(), reactions)
         residual_reactions, stored_dg = zip(*rxn_dg_pairs)
         stored_dg = np.array(stored_dg)
 
-        (standard_dg, dg_sigma) = self.predictor.standard_dg_multi(
-            residual_reactions
+        (standard_dg, dg_uncertainty,) = self.predictor.standard_dg_multi(
+            residual_reactions,
+            uncertainty_representation=uncertainty_representation,
         )
-        return standard_dg + stored_dg, dg_sigma
+
+        standard_dg += Q_(stored_dg, "kJ/mol")
+        return standard_dg, dg_uncertainty
 
     def standard_dg_prime(self, reaction: PhasedReaction) -> ureg.Measurement:
         """Calculate the transformed reaction energies of a reaction.
@@ -543,21 +532,31 @@ class ComponentContribution(object):
         )
 
     def standard_dg_prime_multi(
-        self, reactions: List[PhasedReaction]
-    ) -> Tuple[np.array, np.array]:
+        self,
+        reactions: List[PhasedReaction],
+        uncertainty_representation: str = "cov",
+    ) -> Tuple[Q_, Q_]:
         """Calculate the transformed reaction energies of a list of reactions.
 
+        Parameters
+        ----------
+        reactions : List[PhasedReaction]
+            a list of PhasedReaction objects to estimate
+        uncertainty_representation : str
+            which representation to use for the uncertainties. 'cov' would
+            return a full covariance matrix. 'sqrt' would return a sqaure root
+            of the covariance, based on the uncertainty vectors.
+            'fullrank' would return a full-rank square root of the covariance
+            which is a compressed form of the 'sqrt' result.
+            (Default value: 'cov')
+
         Returns
         -------
-        standard_dg_primes : ndarray
-            a 1D NumPy array containing the CC estimates for the reactions'
-            transformed dG0
-        dg_sigma : ndarray
-            the second is a 2D numpy array containing the covariance matrix
-            of the standard errors of the estimates. one can use the
-            eigenvectors of the matrix to define a confidence high-dimensional
-            space, or use dg_sigma as the covariance of a Gaussian used for
-            sampling (where 'standard_dg_primes' is the mean of that Gaussian).
+        standard_dg : Quantity
+            the CC estimation of the reactions' standard transformed energies
+        dg_uncertainty : Quantity
+            the uncertainty co-variance matrix
+            (in either 'cov', 'sqrt' or 'fullrank' format)
         """
         rxn_dg_pairs = map(
             lambda r: r.separate_stored_dg_prime(
@@ -571,14 +570,30 @@ class ComponentContribution(object):
         residual_reactions, stored_dg_primes = zip(*rxn_dg_pairs)
         stored_dg_primes = np.array(stored_dg_primes)
 
-        (standard_dg_prime, dg_sigma) = self.predictor.standard_dg_prime_multi(
+        (standard_dg_prime, dg_uncertainty,) = self.predictor.standard_dg_multi(
             residual_reactions,
-            p_h=self.p_h,
-            p_mg=self.p_mg,
-            ionic_strength=self.ionic_strength,
-            temperature=self.temperature,
+            uncertainty_representation=uncertainty_representation,
         )
-        return standard_dg_prime + stored_dg_primes, dg_sigma
+
+        # So far, standard_dg_prime is actually storing only the untransformed
+        # standard ΔG. We must add the Legendre transform to each reaction to
+        # get the standard ΔG.
+        for i, r in enumerate(residual_reactions):
+            try:
+                standard_dg_prime[i] += r.transform(
+                    p_h=self.p_h,
+                    p_mg=self.p_mg,
+                    ionic_strength=self.ionic_strength,
+                    temperature=self.temperature,
+                )
+            except MissingDissociationConstantsException as e:
+                warnings.warn(
+                    f"Cannot calculate Legendre transform for reaction #{i}: "
+                    + str(e)
+                )
+
+        standard_dg_prime += Q_(stored_dg_primes, "kJ/mol")
+        return standard_dg_prime, dg_uncertainty
 
     def physiological_dg_prime(
         self, reaction: PhasedReaction
@@ -833,3 +848,86 @@ class ComponentContribution(object):
             self.ccache.is_water,
             self.ccache.water,
         )
+
+    # DISCLAIMER: this is experimental code!
+    def _standard_dg_prime_multi_adjust_dof(
+        self,
+        reactions: List[PhasedReaction],
+        uncertainty_representation: str = "cov",
+    ) -> Tuple[np.array, np.array]:
+        """Calculate the transformed reaction energies of a list of reactions.
+
+        Adjusts the estimates by using the Degrees Of Freedom
+        associated with the CC nullspace to minimize the norm2 of the
+        standard reaction energies - using Least Squares.
+
+
+        Parameters
+        ----------
+        reactions : List[PhasedReaction]
+            a list of PhasedReaction objects to estimate
+        uncertainty_representation : str
+            which representation to use for the uncertainties. 'cov' would
+            return a full covariance matrix. 'sqrt' would return a sqaure root
+            of the covariance, based on the uncertainty vectors.
+            'fullrank' would return a full-rank square root of the covariance
+            which is a compressed form of the 'sqrt' result.
+            (Default value: 'cov')
+
+        Returns
+        -------
+        standard_dg : Quantity
+            the CC estimation of the reactions' standard transformed energies
+        dg_uncertainty : Quantity
+            the uncertainty co-variance matrix
+            (in either 'cov', 'sqrt' or 'fullrank' format)
+        """
+        rxn_dg_pairs = map(
+            lambda r: r.separate_stored_dg_prime(
+                p_h=self.p_h,
+                p_mg=self.p_mg,
+                ionic_strength=self.ionic_strength,
+                temperature=self.temperature,
+            ),
+            reactions,
+        )
+        residual_reactions, stored_dg_primes = zip(*rxn_dg_pairs)
+
+        (standard_dg_prime, dg_uncertainty,) = self.predictor.standard_dg_multi(
+            residual_reactions,
+            uncertainty_representation=uncertainty_representation,
+        )
+
+        # So far, standard_dg_prime is actually storing only the untransformed
+        # standard ΔG. We must add the Legendre transform to each reaction to
+        # get the standard ΔG.
+        for i, r in enumerate(residual_reactions):
+            try:
+                standard_dg_prime[i] += r.transform(
+                    p_h=self.p_h,
+                    p_mg=self.p_mg,
+                    ionic_strength=self.ionic_strength,
+                    temperature=self.temperature,
+                )
+            except MissingDissociationConstantsException as e:
+                warnings.warn(
+                    f"Cannot calculate Legendre transform for reaction #{i}: "
+                    + str(e)
+                )
+
+        standard_dg_prime += Q_(stored_dg_primes, "kJ/mol")
+
+        # Calculate the residual matrix (X) that represents the DOFs that are
+        # completely free (infinite uncertianty).
+        X = self.predictor.preprocess.get_reaction_prediction_orthogonal_dof(
+            residual_reactions
+        )
+
+        # Project the 'standard_dg_prime' onto the nullspace of the residual
+        # matrix (using an orthogonal projection). This doesn't change the
+        # values in directions that are not complete DOFs, and at the same
+        # time minimizes the norm2 of the estimate vector.
+        _, _, _, P_N = LINALG.invert_project(X.T)
+        standard_dg_prime = P_N @ standard_dg_prime
+
+        return standard_dg_prime, dg_uncertainty
