@@ -31,7 +31,7 @@ from typing import Tuple
 
 import numpy as np
 from equilibrator_cache import PROTON_INCHI_KEY, Compound
-from equilibrator_cache.thermodynamic_constants import legendre_transform
+from equilibrator_cache.thermodynamic_constants import R, _legendre_transform
 
 from . import Q_, ureg
 
@@ -352,13 +352,14 @@ class PhasedCompound(object):
         if ms is not None:
             # for compound-phase pairs in this dictionary use the stored
             # value and transform directly using the Legendre transform
-            return ms.standard_dgf + legendre_transform(
-                p_h=p_h,
-                ionic_strength=ionic_strength,
-                temperature=temperature,
-                p_mg=p_mg,
-                num_protons=ms.num_protons,
+            return ms.standard_dgf + R * temperature * _legendre_transform(
+                pH=p_h.m_as(""),
+                pMg=p_mg.m_as(""),
+                ionic_strength_M=ionic_strength.m_as("M"),
+                T_in_K=temperature.m_as("K"),
                 charge=ms.charge,
+                num_protons=ms.num_protons,
+                num_magnesiums=0.0,
             )
         else:
             # for all other compounds, we will use component_contribution
